"""
Methods for Exercises 2
Spring School on Scientific Computing

"""

import numpy as np


def newton(f,df,x0,maxiter,tol):
    """
    Solves f(x) = 0 using Newton's method
    Input:
        f: function
        df: derivative of f (also as a function)
        x0: float, starting value
        maxiter: int, maximum iteration number
        tol: float, stopping criterion
    Output:
        x_n: float, solution
        it: number of iterations
        fun_values: Array of function values (Ex 2.1(c))
    """
        
    xn = x0
    fun = f(xn)
    iter = 0
        
    fun_values = np.array([fun]) # Exercise 2.1(c)
    
    
    while np.abs(fun) >= tol:
        
        dfun = df(xn)
        xn = xn - fun/dfun
        fun = f(xn)
        fun_values = np.append(fun_values,fun)
        
        iter += 1
        
        if iter >= maxiter:
            print('Maximum number of Newton iterations reached.')
            print('The function value ist still', fun)
            break
            
        if xn > 1e9 or xn < -1e9:
            print('Newton iteration diverged')
            break
        
       
    return xn, iter, fun_values


def bisection(f,a,b,tol):
    """
    This method solves a system f(x) = 0 
    using bisection
    Input:
        f: function
        a: float, left interval limit 
        b: float, right interval limit
        tol: float, stopping criterion
    Output:
        x_n: float, solution
        it: number of iterations
        fun_values: Array of function values (Ex 2.2(b))
        
    """
        
    
    l,r = a,b
    fun_l, fun_r = f(l), f(r)
    it = 1
    
    if fun_l*fun_r >= 0:
        print('Error in bisection: Function values need to have different signs')
        return
    
    fun_values = np.array([])  # Exercise 2.2 (b)
    
    c = (l+r)/2
    
    while np.abs(f(c)) >= tol:
        
        if fun_l*f(c) < 0:
            r = c
        else:
            l = c
            
        fun_l, fun_r = f(l), f(r)
        c = (l+r)/2
       
        fun_values = np.append(fun_values,f(c))
        
        it += 1
    
    return c, it, fun_values

#%%


def newton_linesearch(f,df,x0,q,maxiter,tol):
    """
    Solves f(x) = 0 using Newton with (simple) line search
    Input:
        f: function 
        df: analytically computed derivative of f (function)
        x0: float, starting value
        maxiter: int, maximum number of iterations
        tol: float, stopping criterion

    Output:
        x_n: float, final iterate
        it:  int, number of iterations
        fun_values: Array of function values
    """
        
    x_n = x0
    xold = x_n
    fun = f(x_n)
    it = 0
    fun_values = np.array([fun])
    
    while  np.abs(fun) >= tol:
        fun = f(x_n)
        dfun = df(x_n)
        lam = 1
        x_n = xold - fun/dfun
        
        # print('fun is ', fun, ' x_n is', x_n)
        
        while np.abs(f(x_n)) > q*np.abs(fun):
            lam = lam*0.5
            x_n = xold - lam*fun/dfun
        
        print('The stepsize lambda in iterate ',it,' is', lam)
        fun = f(x_n)
        xold = x_n
        
        fun_values = np.append(fun_values,fun)
        it += 1
        if it >= maxiter:
            break

       
    return x_n, it, fun_values


#%%

def newton_Rn(F,J,x0,maxiter,tol):
    """
    Solves a system F(x) = 0 using Newton's method
    Input:
        f: fhandle 
        J: analytically computed Jacobi matrix (fhandle)
        x0: float, starting value
        maxiter: int, maximum number of iterations
        tol: float, stopping criterion

    Output:
        x_n: float, approximative solution of F(x) = 0
        it:  int, number of iterations
        res_it: Sequence of Euklidean norms ||F(x)||
    """
        
    x_n = x0
    fun = F(x_n)
    it = 0
    res_it = [np.linalg.norm(fun)]
    
    while res_it[it] >= tol and it < maxiter:
                
        dfun = J(x_n)
        
        deltaxn = np.linalg.solve(dfun, fun)
        x_n = x_n - deltaxn
        
        fun = F(x_n)
        
        res_it.append(np.linalg.norm(fun))
        it += 1
                
       
    return x_n, it, res_it
